import warnings

from pytest_run_parallel.cpu_detection import get_logical_cpus


def get_configured_num_workers(config):
    n_workers = config.option.parallel_threads
    if n_workers == "auto":
        logical_cpus = get_logical_cpus()
        n_workers = logical_cpus if logical_cpus is not None else 1
    else:
        n_workers = int(n_workers)
    return n_workers


def auto_or_int(val):
    if val == "auto":
        logical_cpus = get_logical_cpus()
        return logical_cpus if logical_cpus is not None else 1
    return int(val)


def get_num_workers(item):
    n_workers = get_configured_num_workers(item.config)
    marker_used = False
    marker = item.get_closest_marker("parallel_threads")
    if marker is not None:
        marker_used = True
        n_workers = auto_or_int(marker.args[0])
        if n_workers > 1:
            warnings.warn(
                "Using the parallel_threads marker with a value greater than 1 is deprecated. "
                "Use parallel_threads_limit if you want to set a thread limit or "
                "force_parallel_threads if you want to force a test to be multithreaded.",
                DeprecationWarning,
                stacklevel=2,
            )
    limit_marker = item.get_closest_marker("parallel_threads_limit")
    if limit_marker is not None:
        val = auto_or_int(limit_marker.args[0])
        if val == 1:
            marker_used = True
        if n_workers > val:
            n_workers = val

    force_marker = item.get_closest_marker("force_parallel_threads")
    if force_marker is not None:
        n_workers = auto_or_int(force_marker.args[0])
        if n_workers == 1:
            marker_used = True

    return n_workers, marker_used


def get_num_iterations(item):
    n_iterations = item.config.option.iterations
    marker = item.get_closest_marker("iterations")
    if marker is not None:
        n_iterations = int(marker.args[0])
    return n_iterations, marker is not None



import ast
import functools
import inspect
import sys
import traceback
import warnings

try:
    # added in hypothesis 6.131.0
    from hypothesis import is_hypothesis_test
except ImportError:
    try:
        # hypothesis versions < 6.131.0
        from hypothesis.internal.detection import is_hypothesis_test
    except ImportError:
        # hypothesis isn't installed
        def is_hypothesis_test(fn):
            return False


try:
    from hypothesis import __version_info__ as hypothesis_version
except ImportError:
    hypothesis_version = (0, 0, 0)

HYPOTHESIS_THREADSAFE_VERSION = (6, 136, 3)

WARNINGS_IS_THREADSAFE = bool(
    getattr(sys.flags, "context_aware_warnings", 0)
    and getattr(sys.flags, "thread_inherit_context", 0)
)

CTYPES_IS_THREADSAFE = sys.version_info > (3, 13)


def construct_base_blocklist(unsafe_warnings, unsafe_ctypes):
    safe_warnings = not unsafe_warnings and WARNINGS_IS_THREADSAFE
    safe_ctypes = not unsafe_ctypes and CTYPES_IS_THREADSAFE
    return {
        ("pytest", "warns", safe_warnings),
        ("pytest", "deprecated_call", safe_warnings),
        ("_pytest.recwarn", "warns", safe_warnings),
        ("_pytest.recwarn", "deprecated_call", safe_warnings),
        ("warnings", "catch_warnings", safe_warnings),
        ("unittest.mock", "*", False),
        ("mock", "*", False),
        ("ctypes", "*", safe_ctypes),
        ("gc", "collect", False),
    }


THREAD_UNSAFE_FIXTURES = {
    "capsys": False,
    "capteesys": False,
    "capsysbinary": False,
    "capfd": False,
    "capfdbinary": False,
    "monkeypatch": False,
    "recwarn": WARNINGS_IS_THREADSAFE,
}


class ThreadUnsafeNodeVisitor(ast.NodeVisitor):
    def __init__(
        self, fn, skip_set, unsafe_warnings, unsafe_ctypes, unsafe_hypothesis, level=0
    ):
        self.thread_unsafe = False
        self.thread_unsafe_reason = None
        blocklist = construct_base_blocklist(unsafe_warnings, unsafe_ctypes)
        self.blocklist = {b[:2] for b in blocklist if not b[-1]} | skip_set
        self.module_blocklist = {mod for mod, func in self.blocklist if func == "*"}
        self.function_blocklist = {
            (mod, func) for mod, func in self.blocklist if func != "*"
        }

        modules = {mod.split(".")[0] for mod, _ in self.blocklist}
        modules |= {mod for mod, _ in self.blocklist}

        self.fn = fn
        self.skip_set = skip_set
        self.unsafe_warnings = unsafe_warnings
        self.unsafe_ctypes = unsafe_ctypes
        self.unsafe_hypothesis = unsafe_hypothesis
        self.level = level
        self.modules_aliases = {}
        self.func_aliases = {}
        self.globals = getattr(fn, "__globals__", {})

        # see issue #121, sometimes __globals__ isn't iterable
        try:
            iter(self.globals)
        except TypeError:
            self.globals = {}

        for var_name in iter(self.globals):
            value = fn.__globals__[var_name]
            if inspect.ismodule(value) and value.__name__ in modules:
                self.modules_aliases[var_name] = value.__name__
            elif inspect.isfunction(value):
                if value.__module__ is None:
                    continue
                if value.__module__ in modules:
                    self.func_aliases[var_name] = (value.__module__, value.__name__)
                    continue

                all_parents = self._create_all_parent_modules(value.__module__)
                for parent in all_parents:
                    if parent in modules:
                        self.func_aliases[var_name] = (parent, value.__name__)
                        break

        super().__init__()

    def _create_all_parent_modules(self, module_name):
        all_parent_modules = set()
        parent, dot, _ = module_name.rpartition(".")
        while dot:
            all_parent_modules.add(parent)
            parent, dot, _ = parent.rpartition(".")
        return all_parent_modules

    def _is_module_blocklisted(self, module_name):
        # fast path
        if module_name in self.module_blocklist:
            return True

        # try parent modules
        all_parents = self._create_all_parent_modules(module_name)
        if any(parent in self.module_blocklist for parent in all_parents):
            return True
        return False

    def _is_function_blocklisted(self, module_name, func_name):
        # Whole module is blocked
        if self._is_module_blocklisted(module_name):
            return True

        # Function is blocked
        if (module_name, func_name) in self.function_blocklist:
            return True

        return False

    def _recursive_analyze_attribute(self, node):
        current = node
        while isinstance(current.value, ast.Attribute):
            current = current.value
        if not isinstance(current.value, ast.Name):
            return
        id = current.value.id

        def _get_child_fn(mod, node):
            if isinstance(node.value, ast.Attribute):
                submod = _get_child_fn(mod, node.value)
                return getattr(submod, node.attr, None)

            if not isinstance(node.value, ast.Name):
                return None
            return getattr(mod, node.attr, None)

        if id in self.globals:
            mod = self.fn.__globals__[id]
            child_fn = _get_child_fn(mod, node)
            if child_fn is not None and callable(child_fn):
                self.thread_unsafe, self.thread_unsafe_reason = (
                    identify_thread_unsafe_nodes(
                        child_fn,
                        self.skip_set,
                        self.unsafe_warnings,
                        self.unsafe_ctypes,
                        self.unsafe_hypothesis,
                        self.level + 1,
                    )
                )

    def _build_attribute_chain(self, node):
        chain = []
        current = node

        while isinstance(current, ast.Attribute):
            chain.insert(0, current.attr)
            current = current.value

        if isinstance(current, ast.Name):
            chain.insert(0, current.id)

        return chain

    def _visit_attribute_call(self, node):
        if isinstance(node.value, ast.Name):
            real_mod = node.value.id
            if real_mod in self.modules_aliases:
                real_mod = self.modules_aliases[real_mod]
            if self._is_function_blocklisted(real_mod, node.attr):
                self.thread_unsafe = True
                self.thread_unsafe_reason = (
                    f"calls thread-unsafe function: {real_mod}.{node.attr}"
                )
            elif self.level < 2:
                self._recursive_analyze_attribute(node)
        elif isinstance(node.value, ast.Attribute):
            chain = self._build_attribute_chain(node)
            module_part = ".".join(chain[:-1])
            func_part = chain[-1]
            if self._is_function_blocklisted(module_part, func_part):
                self.thread_unsafe = True
                self.thread_unsafe_reason = (
                    f"calls thread-unsafe function: {'.'.join(chain)}"
                )
            elif self.level < 2:
                self._recursive_analyze_attribute(node)

    def _recursive_analyze_name(self, node):
        if node.id in self.globals:
            child_fn = self.fn.__globals__[node.id]
            if callable(child_fn):
                self.thread_unsafe, self.thread_unsafe_reason = (
                    identify_thread_unsafe_nodes(
                        child_fn,
                        self.skip_set,
                        self.unsafe_warnings,
                        self.unsafe_ctypes,
                        self.unsafe_hypothesis,
                        self.level + 1,
                    )
                )

    def _visit_name_call(self, node):
        if node.id in self.func_aliases:
            if self._is_function_blocklisted(*self.func_aliases[node.id]):
                self.thread_unsafe = True
                self.thread_unsafe_reason = f"calls thread-unsafe function: {node.id}"
                return

        if self.level < 2:
            self._recursive_analyze_name(node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self._visit_attribute_call(node.func)
        elif isinstance(node.func, ast.Name):
            self._visit_name_call(node.func)
        self.generic_visit(node)

    def visit_Assign(self, node):
        if len(node.targets) == 1:
            name_node = node.targets[0]
            value_node = node.value
            if getattr(name_node, "id", None) == "__thread_safe__" and not bool(
                value_node.value
            ):
                self.thread_unsafe = True
                self.thread_unsafe_reason = (
                    f"calls thread-unsafe function: {self.fn.__name__} "
                    "(inferred via func.__thread_safe__ == False)"
                )
                return

        self.generic_visit(node)

    def visit(self, node):
        if self.thread_unsafe:
            return
        return super().visit(node)


def _is_source_indented(src):
    # Find first nonblank line. If one can't be found, use placeholder.
    non_blank_lines = (line for line in src.split("\n") if line.strip() != "")
    first_non_blank_line = next(non_blank_lines, "pass")
    is_indented = first_non_blank_line[0].isspace()
    return is_indented


def _visit_node(visitor, fn):
    try:
        src = inspect.getsource(fn)
    except (OSError, TypeError):
        # if we can't get the source code (e.g. builtin function) then give up
        # and don't attempt detection but default to assuming thread safety
        return False, None
    if _is_source_indented(src):
        # This test was extracted from a class or indented area, and Python
        # needs to be told to expect indentation.
        src = "if True:\n" + src
    try:
        tree = ast.parse(src)
    except (SyntaxError, ValueError):
        # AST parsing failed because the AST is invalid. Who knows why but that
        # means we can't run thread safety detection. Bail and assume
        # thread-safe.
        return False, None
    visitor.visit(tree)


def _identify_thread_unsafe_nodes(
    fn, skip_set, unsafe_warnings, unsafe_ctypes, unsafe_hypothesis, level=0
):
    if is_hypothesis_test(fn):
        if hypothesis_version < HYPOTHESIS_THREADSAFE_VERSION:
            return (
                True,
                f"uses hypothesis v{'.'.join(map(str, hypothesis_version))}, which "
                "is before the first thread-safe version "
                f"(v{'.'.join(map(str, HYPOTHESIS_THREADSAFE_VERSION))})",
            )
        if unsafe_hypothesis:
            return (
                True,
                "uses Hypothesis, and pytest-run-parallel was run with "
                "--mark-hypothesis-as-unsafe",
            )

    try:
        visitor = ThreadUnsafeNodeVisitor(
            fn, skip_set, unsafe_warnings, unsafe_ctypes, unsafe_hypothesis, level=level
        )
        with warnings.catch_warnings():
            # in case pytest is configured to treat warnings as errors.
            warnings.simplefilter("default")
            _visit_node(visitor, fn)
    except Exception as e:
        tb = traceback.format_exc()
        msg = (
            f"Uncaught exception while checking test '{fn}' for thread-unsafe "
            "functionality. Please report a bug to pytest-run-parallel at "
            "https://github.com/Quansight-Labs/pytest-run-parallel/issues/new "
            "including this message if thread safety detection should work.\n"
            f"{e}\n{tb}\n"
            "Assuming this test is thread-safe."
        )
        warnings.warn(msg, RuntimeWarning)
        return False, None

    return visitor.thread_unsafe, visitor.thread_unsafe_reason


cached_thread_unsafe_identify = functools.lru_cache(_identify_thread_unsafe_nodes)


def identify_thread_unsafe_nodes(*args, **kwargs):
    try:
        return cached_thread_unsafe_identify(*args, **kwargs)
    except TypeError:
        return _identify_thread_unsafe_nodes(*args, **kwargs)


def construct_thread_unsafe_fixtures(config):
    unsafe_fixtures = THREAD_UNSAFE_FIXTURES.copy()
    for item in config.getini("thread_unsafe_fixtures"):
        unsafe_fixtures[item] = False

    if config.option.mark_warnings_as_unsafe:
        unsafe_fixtures["recwarn"] = False

    return {uf[0] for uf in unsafe_fixtures.items() if not uf[1]}



import functools
import os
import re
import sys
import threading
import warnings

import _pytest.doctest
import _pytest.outcomes
import pytest

from pytest_run_parallel.thread_comparator import ThreadComparator
from pytest_run_parallel.thread_unsafe_detection import (
    construct_thread_unsafe_fixtures,
    identify_thread_unsafe_nodes,
)
from pytest_run_parallel.utils import (
    get_configured_num_workers,
    get_num_iterations,
    get_num_workers,
)

GIL_WARNING_MESSAGE_CONTENT = re.compile(
    r"The global interpreter lock \(GIL\) has been enabled to load module '(?P<module>[^']*)'"
)

GIL_ENABLED_ERROR_TEXT = (
    "GIL was dynamically re-enabled during test {stage_test} to load module '{module}'. "
    "When running under a free-threaded interpreter with the GIL initially disabled, "
    "the test suite must not cause the GIL to be re-enabled at runtime. Check "
    "for compiled extension modules that do not use the 'Py_mod_gil' slot or the "
    "'PyUnstable_Module_SetGIL' API. Pass --ignore-gil-enabled in your pytest invocation"
    "to ignore this and run the tests anyway."
)


def wrap_function_parallel(fn, n_workers, n_iterations):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        errors = []
        skip = None
        failed = None
        barrier = threading.Barrier(n_workers)
        original_switch = sys.getswitchinterval()
        new_switch = 1e-6
        for _ in range(3):
            try:
                sys.setswitchinterval(new_switch)
                break
            except ValueError:
                new_switch *= 10
        else:
            sys.setswitchinterval(original_switch)

        try:

            def closure(*args, **kwargs):
                # "smuggling" thread_index into closure with args
                thread_index, args = args[0], args[1:]
                # modifying fixtures
                if n_workers > 1:
                    if "thread_index" in kwargs:
                        kwargs["thread_index"] = thread_index
                    if "tmp_path" in kwargs:
                        kwargs["tmp_path"] = (
                            kwargs["tmp_path"] / f"thread_{thread_index!s}"
                        )
                        kwargs["tmp_path"].mkdir(exist_ok=True)
                    if "tmpdir" in kwargs:
                        kwargs["tmpdir"] = kwargs["tmpdir"].ensure(
                            f"thread_{thread_index!s}", dir=True
                        )

                for i in range(n_iterations):
                    if "iteration_index" in kwargs:
                        kwargs["iteration_index"] = i

                    barrier.wait()
                    try:
                        fn(*args, **kwargs)
                    except Warning:
                        pass
                    except Exception as e:
                        errors.append(e)
                    except _pytest.outcomes.Skipped as s:
                        nonlocal skip
                        skip = s.msg
                    except _pytest.outcomes.Failed as f:
                        nonlocal failed
                        failed = f

            workers = []
            for i in range(0, n_workers):
                worker_kwargs = kwargs
                # "smuggling" i into closure with args to use for thread_index fixture
                workers.append(
                    threading.Thread(
                        target=closure, args=(i, *args), kwargs=worker_kwargs
                    )
                )

            num_completed = 0
            try:
                for worker in workers:
                    worker.start()
                    num_completed += 1
            finally:
                if num_completed < len(workers):
                    barrier.abort()

            for worker in workers:
                worker.join()

            # if we ever want to add cleanup, put it here

        finally:
            sys.setswitchinterval(original_switch)

        if skip is not None:
            pytest.skip(skip)
        elif failed is not None:
            raise failed
        elif errors:
            raise errors[0]

    return inner


class RunParallelPlugin:
    def __init__(self, config):
        self.verbose = bool(int(os.environ.get("PYTEST_RUN_PARALLEL_VERBOSE", "0")))
        self.skip_thread_unsafe = config.option.skip_thread_unsafe
        self.mark_warnings_as_unsafe = config.option.mark_warnings_as_unsafe
        self.mark_ctypes_as_unsafe = config.option.mark_ctypes_as_unsafe
        self.mark_hypothesis_as_unsafe = config.option.mark_hypothesis_as_unsafe
        self.ignore_gil_enabled = config.option.ignore_gil_enabled
        if get_configured_num_workers(config) == 1:
            self.ignore_gil_enabled = 1
        self.forever = config.option.forever

        skipped_functions = [
            x.split(".") for x in config.getini("thread_unsafe_functions")
        ]
        self.skipped_functions = frozenset(
            (".".join(x[:-1]), x[-1]) for x in skipped_functions
        )

        self.unsafe_fixtures = construct_thread_unsafe_fixtures(config)
        self.thread_unsafe = {}
        self.run_in_parallel = {}

    def skipped_or_not_parallel(self, *, plural):
        if plural:
            skipped = "were skipped"
            parallel = "were not run in parallel"
        else:
            skipped = "was skipped"
            parallel = "was not run in parallel"
        return skipped if self.skip_thread_unsafe else parallel

    def _mark_test_thread_unsafe(self, item, reason):
        if self.skip_thread_unsafe:
            item.add_marker(pytest.mark.skip(reason=reason))
        else:
            item.add_marker(pytest.mark.parallel_threads(1))
        self.thread_unsafe[item.nodeid] = reason

    def _is_thread_unsafe(self, item):
        m = item.get_closest_marker("thread_unsafe")
        if m is not None:
            reason = m.kwargs.get("reason", None)
            if reason is None:
                reason = "uses the thread_unsafe marker"
            return True, reason

        fixtures = getattr(item, "fixturenames", ())
        if any(fixture in fixtures for fixture in self.unsafe_fixtures):
            used_unsafe_fixtures = self.unsafe_fixtures & set(fixtures)
            return True, f"uses thread-unsafe fixture(s): {used_unsafe_fixtures}"

        return identify_thread_unsafe_nodes(
            item.obj,
            self.skipped_functions,
            self.mark_warnings_as_unsafe,
            self.mark_ctypes_as_unsafe,
            self.mark_hypothesis_as_unsafe,
        )

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtestloop(self, session: pytest.Session):
        """
        Based on the default implementation in pytest, but also adds support
        for running the tests in an endless loop.
        """
        if not self.forever:
            # let the default pytest_runtestloop run if we don't need any
            # customization for --forever.
            return None

        if (
            session.testsfailed
            and not session.config.option.continue_on_collection_errors
        ):
            raise session.Interrupted(
                "%d errors during collection" % session.testsfailed
            )

        if session.config.option.collectonly:
            return True

        number_of_items = len(session.items)
        if number_of_items == 0:
            raise pytest.UsageError(
                "Test collection found zero tests when passing --forever. Are you sure you searched for the correct tests?"
            )
        iter_number = 0
        idx = 0
        next_idx = (idx + 1) % number_of_items

        while idx < number_of_items:
            if idx == 0:
                print("\n\n", end="")
                print("==========================================================")
                print("You ran the test suite with 'forever' mode enabled.")
                print(f"Running the tests again. This is iteration #{iter_number}.")
                print("==========================================================")
                iter_number += 1

            item = session.items[idx]
            nextitem = session.items[next_idx] if next_idx < number_of_items else None

            item.config.hook.pytest_runtest_protocol(item=item, nextitem=nextitem)
            if session.shouldfail:
                raise session.Failed(session.shouldfail)
            if session.shouldstop:
                raise session.Interrupted(session.shouldstop)

            idx = next_idx
            next_idx = (idx + 1) % number_of_items

        return True

    # This is tryfirst=True because we need our plugin's pytest_collection_finish
    # to be called before the terminal plugin's pytest_collection_finish:
    #
    # - pytest's terminal plugin hooks pytest_collection_finish, which calls
    #   pytest_report_collectionfinish
    # - we hook pytest_report_collectionfinish, assuming _handle_collected_item
    #   has already been called for all items when pytest_report_collectionfinish
    #   is called
    @pytest.hookimpl(tryfirst=True)
    def pytest_collection_finish(self, session):
        for item in session.items:
            self._handle_collected_item(item)

    def _handle_collected_item(self, item):
        if isinstance(item, _pytest.doctest.DoctestItem):
            self._mark_test_thread_unsafe(
                item, "is a doctest (pytest-run-parallel does not support doctests)"
            )
            return

        if getattr(item, "obj", None) is None:
            if not hasattr(item, "_parallel_custom_item"):
                warnings.warn(
                    f"Encountered pytest item with type {type(item)} with no 'obj' "
                    "attribute, which is incompatible with pytest-run-parallel. "
                    f"Tests using {type(item)} will not run in a thread pool.\n"
                    "The pytest-run-parallel plugin only supports custom collection "
                    "tree objects that wrap Python functions stored in an attribute "
                    "named 'obj'.\n"
                    "Define a '_parallel_custom_item' attribute on the pytest item"
                    "instance or class to silence this warning.\n"
                    "If you do not want to use pytest-run-parallel, uninstall it from "
                    "your environment."
                )
            self._mark_test_thread_unsafe(
                item, "is incompatible with pytest-run-parallel"
            )
            return

        n_workers, parallel_threads_marker_used = get_num_workers(item)
        if n_workers < 0:
            raise ValueError("parallel-threads cannot be negative")

        n_iterations, _ = get_num_iterations(item)
        if n_iterations < 0:
            raise ValueError("parallel-threads cannot be negative")

        if n_workers == 1 and parallel_threads_marker_used:
            self._mark_test_thread_unsafe(item, "test is marked as single-threaded")

        if n_workers > 1:
            thread_unsafe, reason = self._is_thread_unsafe(item)
            if thread_unsafe:
                n_workers = 1
                self._mark_test_thread_unsafe(item, reason)
            else:
                self.run_in_parallel[item.nodeid] = n_workers

        if n_workers > 1 or n_iterations > 1:
            original_globals = item.obj.__globals__
            item.obj = wrap_function_parallel(item.obj, n_workers, n_iterations)
            for name in original_globals:
                if name not in item.obj.__globals__:
                    item.obj.__globals__[name] = original_globals[name]

    @pytest.hookimpl(trylast=True)
    def pytest_report_collectionfinish(self, config, start_path, items):
        return f"Collected {len(self.run_in_parallel)} items to run in parallel"

    @pytest.hookimpl(tryfirst=True, wrapper=True)
    def pytest_report_teststatus(self, report, config):
        outcome = yield
        if getattr(report, "when", None) != "call" or getattr(
            report, "keywords", {}
        ).get("xfail", None):
            return outcome

        if report.nodeid in self.run_in_parallel:
            if report.outcome == "passed":
                return "passed", "·", "PARALLEL PASSED"
            if report.outcome == "failed":
                return "error", "e", "PARALLEL FAILED"
        else:
            reason = self.thread_unsafe.get(report.nodeid)
            if reason is not None:
                if report.outcome == "passed":
                    return (
                        "passed",
                        ".",
                        f"PASSED [thread-unsafe]: {reason}",
                    )
                if report.outcome == "failed":
                    return (
                        "failed",
                        "x",
                        f"FAILED ([thread-unsafe]: {reason})",
                    )
        return outcome

    def _write_reasons_summary(self, terminalreporter):
        for nodeid, reason in self.thread_unsafe.items():
            if reason is not None:
                terminalreporter.line(
                    f"{nodeid} {self.skipped_or_not_parallel(plural=False)} because it {reason}"
                )
            else:
                terminalreporter.line(nodeid)

    @pytest.hookimpl(trylast=True)
    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        enabled = get_configured_num_workers(config) > 1
        if not enabled:
            return

        terminalreporter.section("pytest-run-parallel report", "*")

        if self.verbose and self.thread_unsafe:
            self._write_reasons_summary(terminalreporter)
        elif self.thread_unsafe:
            num = len(self.thread_unsafe)
            if num > 1:
                test = "tests"
            else:
                test = "test"
            terminalreporter.line(
                f"{num} {test} {self.skipped_or_not_parallel(plural=num > 1)}"
                " because of use of thread-unsafe functionality, "
                f"to list the tests that {self.skipped_or_not_parallel(plural=True)}, re-run "
                "while setting PYTEST_RUN_PARALLEL_VERBOSE=1 "
                "in your shell environment"
            )
        else:
            terminalreporter.line("All tests were run in parallel! 🎉")

    @pytest.hookimpl(tryfirst=True)
    def pytest_warning_recorded(
        self, warning_message: warnings.WarningMessage, when, nodeid, location
    ):
        mo = re.match(GIL_WARNING_MESSAGE_CONTENT, str(warning_message.message))
        if mo is None or self.ignore_gil_enabled:
            return

        if when == "collect":
            stage = "collection"
        elif when == "runtest":
            stage = "execution"
        else:
            stage = "configuration"
        stage_test = stage
        if nodeid:
            stage_test += f" of '{nodeid}'"
        pytest.exit(
            reason=GIL_ENABLED_ERROR_TEXT.format(
                stage_test=stage_test, module=mo.group("module")
            ),
            returncode=1,
        )


@pytest.fixture
def num_parallel_threads(request):
    return get_num_workers(request.node)[0]


@pytest.fixture
def num_iterations(request):
    return get_num_iterations(request.node)[0]


# overwritten by wrap_function_parallel when using multiple threads
@pytest.fixture
def thread_index():
    return 0


# overwritten by wrap_function_parallel when using multiple iterations
@pytest.fixture
def iteration_index():
    return 0


@pytest.fixture
def thread_comp(num_parallel_threads):
    return ThreadComparator(num_parallel_threads)


def pytest_configure(config):
    if (
        config.option.forever
        and (n := getattr(config.option, "numprocesses", None)) is not None
    ):
        raise pytest.UsageError(
            f"--forever from pytest-run-parallel is incompatible with `-n {n}` from pytest-xdist."
        )

    config.addinivalue_line(
        "markers",
        "parallel_threads(n): run the given test function in parallel "
        "using `n` threads. Deprecated if `n` is greater than 1. Use either "
        "the parallel_threads_limit or force_parallel_threads marker to make "
        "your intention clearer.",
    )
    config.addinivalue_line(
        "markers",
        "parallel_threads_limit(n): run the given test function in parallel "
        "using a maximum of `n` threads.",
    )
    config.addinivalue_line(
        "markers",
        "force_parallel_threads(n): force the given test function to run in "
        "parallel using `n` threads, overriding the --parallel-threads "
        "command-line argument and any parallel_threads_limit marker. Use "
        "this to ensure a test always runs with a specific number of threads.",
    )
    config.addinivalue_line(
        "markers",
        "iterations(n): run the given test function `n` times in each thread",
    )
    config.addinivalue_line(
        "markers",
        "thread_unsafe: mark the test function as single-threaded",
    )
    config.pluginmanager.register(RunParallelPlugin(config), "_run-parallel")


def pytest_addoption(parser):
    # Note: new options should be on group, not parser
    group = parser.getgroup("run-parallel")
    group.addoption(
        "--parallel-threads",
        action="store",
        dest="parallel_threads",
        default=1,
        help="Set the number of threads used to execute each test concurrently. (default: "
        "%(default)s)",
    )
    group.addoption(
        "--iterations",
        action="store",
        dest="iterations",
        default=1,
        type=int,
        help="Set the number of iterations that each thread will run. (default: %(default)s)",
    )
    group.addoption(
        "--skip-thread-unsafe",
        action="store",
        dest="skip_thread_unsafe",
        help="Whether to skip running thread-unsafe tests. If not provided, thread-unsafe tests "
        "will still run, but only in one thread.",
        type=bool,
        default=False,
    )
    group.addoption(
        "--mark-warnings-as-unsafe",
        action="store_true",
        dest="mark_warnings_as_unsafe",
        default=False,
        help="Mark warnings capture, such as pytest.warns(), as thread-unsafe. If not provided, "
        "the thread safety of warnings capture will be determined automatically.",
    )
    group.addoption(
        "--mark-ctypes-as-unsafe",
        action="store_true",
        dest="mark_ctypes_as_unsafe",
        default=False,
        help="Mark all uses of ctypes as thread-unsafe. If not provided, the thread safety of "
        "ctypes (but not the underlying C code) will be determined automatically.",
    )
    group.addoption(
        "--mark-hypothesis-as-unsafe",
        action="store_true",
        dest="mark_hypothesis_as_unsafe",
        default=False,
        help="Mark hypothesis as thread-unsafe. If not provided, the thread safety of hypothesis "
        "will be determined automatically.",
    )
    group.addoption(
        "--ignore-gil-enabled",
        action="store_true",
        dest="ignore_gil_enabled",
        default=False,
        help="Ignore the GIL becoming enabled in the middle of a test. By default, if the GIL is "
        "re-enabled at runtime, pytest will exit with a non-zero exit code. This option has no "
        "effect for non-free-threaded builds.",
    )
    group.addoption(
        "--forever",
        action="store_true",
        dest="forever",
        default=False,
        help="Run the test loop forever (starting from the top when all the tests have been run), "
        "until one crashes or the user explicitly stops the process with Ctrl-C. This is especially "
        "helpful for hitting thread safety bugs that only occur rarely. --forever is not compatible "
        "with -n from pytest-xdist.",
    )
    parser.addini(
        "thread_unsafe_fixtures",
        "list of thread-unsafe fixture names that cause a test to be run sequentially",
        type="linelist",
        default=[],
    )
    parser.addini(
        "thread_unsafe_functions",
        "list of thread-unsafe fully-qualified named functions that cause "
        "a test to run on one thread",
        type="linelist",
        default=[],
    )
